/*
  * StdFile:		XCMD to position the standard file open dialog box to a particular volume & folder.
  *				Can also be built as an FKEY for use in a QuicKeys environment, where it takes its
  *				argument from a text key (Command-Option-Shift-Zero).
  *				Dewi Williams, Delphi: DEWI.
  */

#include	<MemoryMgr.h>
#include	<EventMgr.h>
#include	<FileMgr.h>
#include	<HFS.h>
#include	<StdFilePkg.h>
#include	<HyperXCmd.h>			/* Defines and structures for HyperCard XCMD programming */
#include	"QuicKeys.h"

/* Build a FKEY instead of a Hypercard XCMD. */
#define	FKEY 	1
#define	NULL		0L

/* Forward references */
QuicInitBlock	*FindSysHeap();
StringPtr		pstrchr(StringPtr str, unsigned char c);
Boolean		FindSFArg(QuicInitBlock *qb, StringPtr arg);

pascal void
main(paramPtr)
#ifndef	FKEY
XCmdBlockPtr	paramPtr;
#endif
{
	DirInfo				dirInfo;
	HVolumeParam			v;
	register unsigned char 	*partialPath;
	register Byte			slen;
	Str255				pathName;
	QuicInitBlock			*qb;

	/* Check for HFS */
	if (FSFCBLen == -1) {						/* Not running HFS */
		SysBeep(1);
		return;
	}

#ifdef	FKEY
	/* Retrieve the path. It's to be found as a program-specific text item (Cmd-Option-Shift-0 (zero)). */
	if ( (qb = FindSysHeap()) == NULL) {
		SysBeep(1);
		return;
	}

	if (FindSFArg(qb, pathName) == FALSE) {
		/* Keycode doesn't exist. */
		SysBeep(1);
		return;
	}
#else
	/* There should only be one parameter. */
	if (paramPtr->paramCount != 1) {
		SysBeep(1);
		return;
	}

	/* First param is the pathname.  Convert it to a pascal string */
	ZeroToPas(paramPtr,*(paramPtr->params[0]), pathName);
#endif

	/* The path  must contain at least 1 colon. */
	if ((partialPath = pstrchr(pathName, ':')) == NULL) {
		SysBeep(1);
		return;
	}

	/* Temporarily isolate the volume component of the pathname. */
	slen = pathName[0];							/* Save the length */
	pathName[0] = (long)partialPath - (long)pathName;

	/* First, get the vRefNum corresponding to Programming: */
	v.ioCompletion = NULL;
	v.ioNamePtr = pathName;
	v.ioVRefNum = 0x8000;		/* See TechNote #77 */
	v.ioVolIndex = -1; 			/* use name & vrefnum */
	
	if (PBHGetVInfo(&v, FALSE) != noErr) {
		SysBeep(1);
		return;
	}

	pathName[0] = slen;			/* Restore the full path */

	/* Now we need the dirID corresponding to the partial path. */
	dirInfo.ioCompletion = NULL;
	dirInfo.ioNamePtr = pathName;
	dirInfo.ioVRefNum = 0;
	dirInfo.ioFDirIndex = 0;		/* use name and vrefnum */
	dirInfo.ioDrDirID = 0;

	if (PBGetCatInfo(&dirInfo, FALSE) != noErr) {
		SysBeep(1);
		return;
	}

	/* And is it a folder? */
	if ( ! (dirInfo.ioFlAttrib & 0x10 )) {		/* It's a file! */
		SysBeep(1);
		return;
	}

	/* Set up the folder info for StdFile. */
	CurDirStore = dirInfo.ioDrDirID;

	/* Set up the volume info for StdFile. */
	SFSaveDisk = -1 * v.ioVRefNum;
}

/*
  * pstrchr:		return pointer to first occurrence of char in pstring.
  */

StringPtr
pstrchr(str, c)
register StringPtr		str;
register unsigned char	c;
{
	register StringPtr	end = str + str[0] + 1;
	
	for (str++;str < end; str++)
		if (*str == c) return str;

	return NULL;
}

#ifdef FKEY
/*
  * This is a  C version of the example code in Chapter 8 of the QuicKeys manual.
  */

QuicInitBlock *
FindSysHeap()
{
	register Ptr			endBlk = SysZone->bkLim;
	register QuicInitBlock	*qp;
	
	qp = (QuicInitBlock *) &SysZone->heapData;

	while(qp != (QuicInitBlock *) endBlk) {
		/* Analyze the block we're looking at. */
		if ((qp->header[0] & 0xC0) == 0x40) {				/* Is it non-relocatable? */
			/* Check magic and signature. */
			if (qp->quic.magic == 0xa89f1234 && qp->quic.signature == 'CELN') {
				/* Chapter 8 states that the version number is 1. It's actually 0 (as the example
				  * assembly language shows and a quick call to CE Software confirmed).
				  */
				if (qp->quic.version == 0) return qp;
			}
		}

		/* Time to move on to the next block. The 0xFFFFFF strips off the tag byte(s). */
		qp = (QuicInitBlock *)(((Byte *)qp) + ((* (long *)&qp->header) & 0xFFFFFF));
	}
	return NULL;		/* Failed to find it */
}

/*
  *	FindSFArg:	Hunts for current contents of Cmd-Option-Shift-Zero.
  */

Boolean
FindSFArg(qb, arg)
QuicInitBlock	*qb;
StringPtr		arg;
{
	register KeyRecord	*kr = qb->quic.application;
	register KeyRecord	*end = kr + N_QCKEYS;

	/* 0x1de2 is the keycode/charcode for Cmd-Option-Shift-0 (found by experimentation). */
	for (; kr < end; kr++) {
		if (kr->QKtype == QK_TEXT) {
			if ((kr->key) == 0x1de2 && (kr->modifiers == (cmdKey|optionKey|shiftKey))) {
				BlockMove(kr->u.QuicText.text, arg, (Size)(kr->u.QuicText.text[0] +1));
				return TRUE;
			}
		}
	}
	return FALSE;
}
#endif
